import warnings

from aiohttp.helpers import AccessLogger
from aiohttp.web_server import RequestHandler
from aiokts.web.error import ServerError

from aiokts.util.json_utils import json_dumps
from aiokts.web.response import ApiErrorResponse, KtsResponse


class KtsAccessLogger(AccessLogger):
    def log_request(self, ctx, request, response, time):
        """Log access.

        :param ctx: Requset context
        :param request: Request object.
        :param response: Response object.
        :param float time: Time taken to serve the request.
        """

        try:
            fmt_info = self._format_line(request, response, time)

            values = list()
            extra = dict()
            for key, value in fmt_info:
                values.append(value)

                if key.__class__ is str:
                    extra[key] = value
                else:
                    extra[key[0]] = {key[1]: value}

            ctx.logger.info(self._log_format % tuple(values), extra=extra)
        except Exception:
            ctx.logger.exception("Error in logging")


class KtsRequestHandler(RequestHandler):
    def __init__(self, *args, **kwargs):
        access_log = kwargs.get('access_log')
        access_log_format = kwargs.get('access_log_format')
        super(KtsRequestHandler, self).__init__(*args, **kwargs)

        if access_log:
            self.access_logger = KtsAccessLogger(access_log, access_log_format)

    def log_access(self, request, response, time):
        if self.access_logger:
            if not hasattr(response, 'ctx'):
                self.access_logger.log(request, response, time)
                return

            ctx = response.ctx
            if ctx is None:
                warnings.warn('Response\'s ctx attribute is None')
                self.access_logger.log(request, response, time)
            else:
                self.access_logger.log_request(ctx, request, response, time)

    def handle_error(self, request, status=500, exc=None, message=None):
        """Handle errors.

            Returns HTTP response with specific status code. Logs additional
            information. It always closes current connection."""
        self.log_exception("Error handling request", exc_info=exc)

        if status == 500:
            error = ServerError.INTERNAL_ERROR
            data = ApiErrorResponse.generate_response_dict(
                message=error.message,
                data=error.payload
            )
            msg = json_dumps(data)
            content_type = 'application/json'
        else:
            msg = message
            content_type = 'text/plain'

        resp = KtsResponse(status=status, text=msg, content_type=content_type,
                           charset='utf-8', ctx=request.ctx)
        resp.force_close()

        # some data already got sent, connection is broken
        if request.writer.output_size > 0 or self.transport is None:
            self.force_close()

        return resp



import functools
import json

from aiokts.util.arguments import check_arguments
from aiokts.web.error import ServerError


def arguments_params(arglist=None):
    if arglist is None:
        arglist = {}

    def _arguments(func):
        @functools.wraps(func)
        async def inner(self, *args, **kwargs):
            if self.request.method == 'GET':
                source = self.request.url.query
            else:
                if self.request.content_type.startswith('application/json'):
                    try:
                        source = await self.request.json()
                    except json.JSONDecodeError:
                        raise ServerError(ServerError.BAD_REQUEST(
                            message='Body must be a valid json'))
                else:
                    source = await self.request.post()
            checked_args = check_arguments(arglist, source, cast_type=True)
            kwargs.update(checked_args)
            return await func(self, *args, **kwargs)

        inner._has_arguments_ = True
        inner.arglist = arglist
        return inner

    return _arguments


def arguments_params_get(arglist=None):
    if arglist is None:
        arglist = {}

    def _arguments(func):
        @functools.wraps(func)
        def inner(self, *args, **kwargs):
            checked_args = check_arguments(
                arglist, self.request.url.query, cast_type=True)
            kwargs.update(checked_args)
            return func(self, *args, **kwargs)

        inner._has_arguments_ = True
        inner.arglist = arglist
        return inner

    return _arguments


def arguments_params_post(arglist=None):
    if arglist is None:
        arglist = {}

    def _arguments(func):
        @functools.wraps(func)
        async def inner(self, *args, **kwargs):
            data = await self.request.post()
            checked_args = check_arguments(arglist, data, cast_type=True)
            kwargs.update(checked_args)
            return await func(self, *args, **kwargs)

        inner._has_arguments_ = True
        inner.arglist = arglist
        return inner

    return _arguments


def arguments_params_json(arglist=None):
    if arglist is None:
        arglist = {}

    def _arguments(func):
        @functools.wraps(func)
        async def inner(self, *args, **kwargs):
            try:
                data = await self.request.json()
            except json.JSONDecodeError:
                raise ServerError(ServerError.BAD_REQUEST(
                    message='Body must be a valid json'))
            checked_args = check_arguments(arglist, data, cast_type=True)
            kwargs.update(checked_args)
            return await func(self, *args, **kwargs)

        inner._has_arguments_ = True
        inner.arglist = arglist
        return inner

    return _arguments



import collections

import logging

import datetime

from aiokts.util.json_utils import JsonSerializable


class Field:
    def __init__(self, default, private):
        self.name = None
        self.model = None
        self.default = default
        self.private = private

    def transform_in(self, value):
        return value

    def transform_to_json(self, value):
        return value

    def __call__(self, value):
        return self.transform_in(value)


class StringField(Field):
    def __init__(self, default=None, private=False):
        super().__init__(default, private)

    def transform_in(self, value):
        return str(value)


class IntField(Field):
    def __init__(self, default=None, private=False):
        super().__init__(default, private)

    def transform_in(self, value):
        return int(value)


class BooleanField(Field):
    def __init__(self, default=None, private=False):
        super().__init__(default, private)

    def transform_in(self, value):
        if value == 'true':
            value = True
        elif value == 'false':
            value = False
        return bool(value)


class UnixTimestampField(Field):
    def __init__(self, default=None, private=False):
        super().__init__(default, private)

    def transform_in(self, value):
        value = int(value)
        return datetime.datetime.fromtimestamp(value)


class IntEnumField(Field):
    def __init__(self, enum_cls, default=None, private=False, json_name=False):
        super().__init__(default, private)
        self.enum_cls = enum_cls
        self.json_name = json_name

    def transform_in(self, value):
        return self.enum_cls(value)

    def transform_to_json(self, value):
        if self.json_name:
            return value.name
        return value.value


class DictField(Field):
    def __init__(self, default=None, private=False):
        super().__init__(default, private)

    def transform_in(self, value):
        assert isinstance(value, dict), \
            'value is not dict (but {}) for {}.{}'.format(
                type(value), self.model.__name__, self.name
            )
        return value


class ForeignModelField(Field):
    def __init__(self, model_cls, default=None, private=False):
        super().__init__(default, private)
        self.model_cls = model_cls

    def transform_in(self, value):
        return self.model_cls.parse(value)

    def transform_to_json(self, value):
        return value.__to_json__()


class DoesNotExistBase(Exception):
    MODEL_CLS = None

    def __init__(self, entity_id=None, message=None):
        self.entity_id = entity_id
        if message is None:
            self.message = \
                "Entity of type '{}' with id {} not found".format(
                    self.MODEL_CLS.__name__, self.entity_id)
        else:
            self.message = message

    def __str__(self):
        return self.message


class ModelMetaclass(type):
    @classmethod
    def __prepare__(mcs, name, bases):
        return collections.OrderedDict()

    def __new__(mcs, class_name, bases, class_dict):
        if class_name != 'Model':
            fields = collections.OrderedDict()
            for name, value in class_dict.items():
                if not name.startswith('__') \
                        and isinstance(value, Field):
                    fields[name] = value
            for name in fields:
                del class_dict[name]

            class_dict['_fields'] = fields

        return super().__new__(mcs, class_name, bases, class_dict)

    def __init__(cls, class_name, bases, class_dict):
        class DoesNotExist(DoesNotExistBase):
            MODEL_CLS = cls
        cls.DoesNotExist = DoesNotExist
        fields = class_dict['_fields']
        if fields is not None:
            for name, f in fields.items():
                f.name = name
                f.model = cls
        super().__init__(class_name, bases, class_dict)


class Model(JsonSerializable, metaclass=ModelMetaclass):
    _fields = None
    DoesNotExist = None
    LOGGER = logging.getLogger('aiokts.models')

    def __init__(self, *args, **kwargs):
        i = 0

        used_kwargs = set()
        for name, field in self._fields.items():
            setattr(self, name, field.default)
            if len(args) <= i:
                # supplying kw arguments
                if name in kwargs:
                    v = kwargs[name]
                    if v is not None:
                        try:
                            v = field.transform_in(v)
                        except Exception as e:
                            raise Exception('{} for field `{}` in {}'.format(
                                str(e), name, self.__class__
                            ))
                        transformer = 'transform_{}'.format(name)
                        if hasattr(self, transformer) \
                                and callable(getattr(self, transformer)):
                            v = getattr(self, transformer)(v)
                    setattr(self, name, v)
                    used_kwargs.add(name)
            else:
                # supplying positional arguments
                v = args[i]
                if v is not None:
                    v = field.transform_in(v)
                setattr(self, name, v)
                i += 1

        if len(args) > i:
            self.LOGGER.warning(
                'Too many positional arguments passed. '
                'Expected %s max', len(self._fields)
            )
            return

        extra_kwargs = set(kwargs.keys()) - used_kwargs
        if len(extra_kwargs) > 0:
            self.LOGGER.warning(
                'Unknown fields passed: %s', extra_kwargs)

    def __to_json__(self):
        res = {}
        for name, field in self._fields.items():
            if field.private:
                continue
            v = getattr(self, name)
            if v is not None:
                v = field.transform_to_json(v)
                transformer = 'transform_json_{}'.format(name)
                if hasattr(self, transformer) \
                        and callable(getattr(self, transformer)):
                    v = getattr(self, transformer)(v)
            res[name] = v
        return res

    @classmethod
    def parse(cls, d: dict):
        if d is None:
            return None
        return cls(**d)

    @classmethod
    def parse_list(cls, l: list):
        if l is None or len(l) == 0:
            return []
        return list(map(lambda d: cls(**d), l))



from aiohttp import web
from aiohttp.log import web_logger
from aiokts.web.request import KtsRequest
from aiokts.web.server import KtsServer

from aiokts.web.context import Context

sentinel = object()


class KtsHttpApplication(web.Application):
    ROUTES = []

    def __init__(self, **kwargs):
        kwargs['debug'] = kwargs.get('debug', False)
        super().__init__(**kwargs)

        for route in self.ROUTES:
            method, path, view_cls = route
            self.router.add_route(method, path, view_cls)

    def _make_request(self, message, payload, protocol, writer, task,
                      _cls=KtsRequest):
        req = super()._make_request(message, payload, protocol, writer, task,
                                    _cls=_cls)
        ctx = self.make_context(req)
        ctx.log_request()
        req.set_context(ctx)
        return req

    def make_context(self, request):
        return Context(request)

    def make_server(self, *, server_cls=KtsServer, **kwargs):
        return server_cls(**kwargs)

    def make_handler(self, *, loop=None,
                     secure_proxy_ssl_header=None, **kwargs):
        self._set_loop(loop)
        self.freeze()

        kwargs['debug'] = self.debug
        if self._handler_args:
            for k, v in self._handler_args.items():
                kwargs[k] = v

        self._secure_proxy_ssl_header = secure_proxy_ssl_header
        return self.make_server(handler=self._handle,
                                request_factory=self._make_request,
                                loop=self.loop, **kwargs)



import asyncio
import logging
import weakref

import api_hour

from aiokts.store import Store
from aiokts.web.application import KtsHttpApplication


class KtsContainerHttpApplication(KtsHttpApplication):
    def __init__(self, config, container=None, **kwargs):
        super().__init__(**kwargs)

        self.config = config
        self._container = weakref.ref(container) if container else None

    @property
    def container(self):
        return self._container() if self._container else None


class KtsWebContainer(api_hour.Container):
    ACCESS_LOG_FORMAT = '%a -> %r %s %b [%Tfs]'

    STORE_NEED = None

    def __init__(self, application=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(__name__)
        self._store = self.make_store()
        self._store_connect_coro = None

        self._application = None
        self.application = application

    @property
    def application(self):
        return self._application

    @application.setter
    def application(self, app):
        if app is not None:
            self._application = app
            self.servers['http'] = app

    async def make_servers(self, sockets):
        handlers = {}
        if 'http' in self.servers:
            handler = self.servers['http'].make_handler(
                logger=self.worker.log,
                keep_alive=self.worker.cfg.keepalive,
                access_log=self.worker.log.access_log,
                access_log_format=self.ACCESS_LOG_FORMAT
            )
            for sock in sockets:
                srv = await self.loop.create_server(handler, sock=sock.sock)
                handlers[srv] = handler
        return handlers

    async def start(self):
        # no await - no need to wait when everything is connected
        asyncio.ensure_future(self.store_connect(), loop=self.loop)
        await super().start()

    async def stop(self):
        await self.store_disconnect()
        await super().stop()

    @property
    def store(self):
        return self._store

    def make_store(self):
        return Store(self.config['store'],
                     need=self.STORE_NEED, loop=self.loop)

    async def store_connect(self):
        if self._store is None:
            return

        self.logger.info('Starting store...')

        def on_finish(f):
            self.logger.info('Store is fully connected')
            self._store_connect_coro = None

        self._store_connect_coro = asyncio.ensure_future(
            self.store.connect(),
            loop=self.loop
        )
        self._store_connect_coro.add_done_callback(on_finish)
        return await self._store_connect_coro

    async def store_disconnect(self):
        if self._store is None:
            return
        self.logger.info('Stopping store...')

        if self._store_connect_coro is not None:
            self._store_connect_coro.cancel()
            self._store_connect_coro = None

        def on_finish(f):
            self.logger.info('Store is fully stopped')

        coro = asyncio.ensure_future(
            self.store.disconnect(),
            loop=self.loop
        )
        coro.add_done_callback(on_finish)
        return await coro



import abc
import asyncio
import logging
import logging.config
import weakref

from aiokts.store import Store


class Command:
    def __init__(self, manager=None):
        self.logger = logging.getLogger(
            'CmdLogger_{}'.format(self.__class__.__name__))
        self._manager = weakref.ref(manager) if manager is not None else None

    @property
    def manager(self):
        return self._manager() if self._manager is not None else None

    @property
    def config(self):
        return self.manager.config

    @abc.abstractmethod
    def run(self, *args):
        pass


class AsyncCommand(Command):
    def __init__(self, manager=None, loop=None):
        super().__init__(manager)
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop

    @abc.abstractmethod
    @asyncio.coroutine
    def run_async(self, *args):
        pass

    def run(self, *args):
        self.loop.run_until_complete(self.run_async(*args))
        return 0


class StoreCommand(AsyncCommand):
    WAIT_TIMEOUT = 3.0
    STORE_CONNECT = True
    STORE_NEED = None

    def __init__(self, manager=None, store_config=None, loop=None):
        super().__init__(manager, loop)
        self.store_config = store_config
        if self.store_config is None:
            self.store_config = self.config.get("store", {})
        self.store = self.make_store()

    def make_store(self):
        if self.manager.store_cls is None:
            return None
        return self.manager.store_cls(self.store_config,
                                      need=self.STORE_NEED,
                                      loop=self.loop)

    async def _store_connect(self):
        await asyncio.wait_for(self.store.connect(),
                               self.WAIT_TIMEOUT,
                               loop=self.loop)

    async def _store_disconnect(self):
        await self.store.disconnect()

    def run(self, *args):
        if self.store is not None and self.STORE_CONNECT:
            self.loop.run_until_complete(self._store_connect())
        try:
            result = super().run(*args)
        finally:
            if self.store is not None and self.STORE_CONNECT:
                self.loop.run_until_complete(self._store_disconnect())
        return result



import abc
import logging
import re
import weakref
from asyncio.locks import Event


class ConfigurationError(Exception):
    pass


class BaseAccessorException(Exception):
    pass


class BaseAccessor(object):
    DEFAULT_HOST = '127.0.0.1'
    DEFAULT_PORT = None
    CHECK_CONFIG = True

    DEFAULT_USERNAME = None
    DEFAULT_PASSWORD = None

    def __init__(self, config, type, store, loop=None):
        super(BaseAccessor, self).__init__()
        self.config = config
        self.type = type
        self._store = weakref.ref(store)
        self.loop = loop
        self.logger = logging.getLogger('accessor[{}]'.format(self.type))

        self.connected = False
        self.connecting = False
        self.disconnecting = False
        self._connected_event = Event(loop=self.loop)
        self._host = None
        self._port = None
        self._username = None
        self._password = None

        if self.CHECK_CONFIG:
            self.check_config()

    @property
    def store(self):
        return self._store() if self._store is not None else None

    async def connect(self):
        if not self.connecting:
            self.connecting = True
            if not self.connected:
                self.logger.info('Connecting to db %s', self.fingerprint)
                await self._connect()
                self.connected = True
                self._connected_event.set()
                self.logger.info('Connected to db %s', self.fingerprint)
            self.connecting = False

    async def disconnect(self):
        if not self.disconnecting:
            self.disconnecting = True
            if self.connected:
                self.logger.info('Disconnecting from db %s', self.fingerprint)
                await self._disconnect()
                self.connected = False
                self._connected_event.clear()
                self.logger.info('Disconnected from db %s', self.fingerprint)
            self.disconnecting = False

    def wait_connected(self):
        return self._connected_event.wait()

    @property
    def fingerprint(self):
        return '[{}://{}:{}]'.format(self.type, self.host, self.port)

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def username(self):
        return self._username

    @property
    def password(self):
        return self._password

    @property
    def has_credentials(self):
        return self.username is not None and self.password is not None

    def check_config(self):
        assert self.type is not None

        if self.config is None:
            raise ConfigurationError(
                'config is required for a connector {}'.format(
                    self.__class__.__name__)
            )

        self._username = self.config.get('username', self.DEFAULT_USERNAME)
        self._password = self.config.get('password', self.DEFAULT_PASSWORD)
        self._host = self.config.get('host', self.DEFAULT_HOST)
        self._port = self.config.get('port', self.DEFAULT_PORT)

        if (self.username is not None and self.password is None) or \
                (self.username is None and self.password is not None):
            raise ConfigurationError(
                'Either both username and password should be defined or none')

    @abc.abstractmethod
    async def _connect(self):
        pass

    @abc.abstractmethod
    async def _disconnect(self):
        pass



from aiohttp.web_request import Request


class KtsRequest(Request):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._context = None

    def set_context(self, ctx):
        self._context = ctx

    @property
    def ctx(self):
        return self._context



import abc
import asyncio
import logging
import logging.config

from aiokts.store import Store


class Daemon(object):
    def __init__(self, *, loop=None, **kwargs):
        self.name = self.__class__.__name__
        self.logger = logging.getLogger('Daemon[{}]'.format(self.name))
        self.loop = loop
        self._stopped = False
        self._run_task = None

    @property
    def is_stopped(self):
        return self._stopped

    @property
    def is_running(self):
        return not self._stopped

    @property
    def config(self):
        return {}

    def check(self):
        pass

    async def start(self):
        self.check()
        self.logger.info('Starting daemon %s', self.name)
        try:
            self._run_task = asyncio.ensure_future(self.run(), loop=self.loop)
            await self._run_task
        except asyncio.futures.CancelledError:
            pass
        except Exception as e:
            logging.exception(e)

    async def stop(self):
        self.logger.info('Stopping %s daemon', self.name)
        if asyncio.iscoroutine(self.cancel) \
                or asyncio.iscoroutinefunction(self.cancel):
            await self.cancel()
        else:
            self.cancel()
        if self._run_task is not None:
            self._run_task.cancel()
            self._run_task = None
        self._stopped = True
        logging.info('Stopped %s daemon', self.name)

    @abc.abstractmethod
    async def run(self):
        raise NotImplementedError()

    def cancel(self):
        pass


class BaseStoreDaemon(Daemon):
    STORE_NEED = []

    def __init__(self, store_config=None, **kwargs):
        super().__init__(**kwargs)

        self.store_config = store_config or self.config.get("store", {})
        self.store = self.make_store()
        self._store_connect_coro = None

    def make_store(self):
        return Store(self.store_config,
                     need=self.STORE_NEED,
                     loop=self.loop)

    def _on_store_connected(self, f):
        self.logger.info('Store connected')
        self._store_connect_coro = None

    def start(self):
        if self.store is not None:
            self._store_connect_coro = asyncio.ensure_future(
                self.store.connect(),
                loop=self.loop
            )
            self._store_connect_coro.add_done_callback(
                self._on_store_connected)
        return super().start()

    async def stop(self):
        await super().stop()
        if self.store is not None:
            if self._store_connect_coro:
                self._store_connect_coro.cancel()
                self._store_connect_coro = None
            await self.store.disconnect()
            self.logger.info('Store disconnected')


def main(daemon):
    loop = asyncio.get_event_loop()

    try:
        loop.run_until_complete(daemon.start())
        loop.run_until_complete(daemon.stop())
        return 0
    except KeyboardInterrupt:
        loop.run_until_complete(daemon.stop())
        return 0
    except:
        return 1



import asyncio

import asynctnt

from aiokts.store.base_accessors import BaseAccessor


class BaseTarantoolAccessor(BaseAccessor):
    DEFAULT_PORT = 3301

    def __init__(self, config, type, store, loop=None):
        super().__init__(config, type, store, loop=loop)
        self._conn = asynctnt.Connection(
            host=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
            loop=self.loop
        )

    async def _connect(self):
        await self._conn.connect()

    async def _disconnect(self):
        if self._conn:
            await self._conn.disconnect()

    @property
    def conn(self):
        return self._conn



import asyncio
import asyncio.futures
import weakref

import pymysql
from io import StringIO
from aiomysql.sa import create_engine
import sqlalchemy as sa

from aiokts.store.base_accessors import BaseAccessor
from aiokts.store.base_accessors.mysql_utils import *


class BaseMySQLAccessor(BaseAccessor):
    DEFAULT_PORT = 3306

    def __init__(self, config, type, store, loop=None):
        super().__init__(config, type, store, loop=loop)
        self.engine = None

        self.connect_timeout = float(self.config.get('connect_timeout', 5.0))
        self.reconnect_timeout = float(
            self.config.get('reconnect_timeout', 1.0))
        self.request_timeout = float(self.config.get('request_timeout', 15.0))
        self.trx_commit_timeout = float(
            self.config.get('trx_commit_timeout', 60.0))

    async def _connect(self):
        while True:
            try:
                def mysql_create_engine(self):
                    return create_engine(
                        minsize=int(self.config.get('pool_size', 10)),
                        maxsize=int(self.config.get('pool_size', 10)),
                        host=self.host,
                        port=self.port,
                        user=self.username,
                        password=self.password,
                        db=self.config['db'],
                        charset='utf8mb4',
                        loop=self.loop,
                        autocommit=True
                    )

                def create_engine_finished(f: asyncio.Future):
                    if f.cancelled():
                        return
                    exc = f.exception()  # need to retrieve exception

                mysql_create_engine_coro = asyncio.ensure_future(
                    mysql_create_engine(self), loop=self.loop)
                mysql_create_engine_coro.add_done_callback(
                    create_engine_finished)
                self.engine = await asyncio.wait_for(
                    mysql_create_engine_coro, self.connect_timeout,
                    loop=self.loop)
                if self.engine is None:
                    await asyncio.sleep(self.reconnect_timeout, loop=self.loop)
                    continue
                setattr(self.engine, '_connector_', weakref.ref(self))
                break
            except Exception as e:
                self.logger.error(
                    '%s Cannot connect to MySQL '
                    '(reconnecting in %s seconds): %s|%s',
                    self.fingerprint, self.reconnect_timeout, repr(e), str(e))
                await asyncio.sleep(self.reconnect_timeout, loop=self.loop)

    async def _disconnect(self):
        if self.engine is not None:
            self.engine.close()
            await self.engine.wait_closed()

    async def ping(self):
        try:
            async with self.acquire() as conn:
                ping = await conn.connection.ping()
        except Exception as e:
            ping = False
        if ping is None:
            ping = True
        return ping

    @property
    def dsn(self):
        return 'mysql+pymysql://{}:{}@{}:{}/{}'.format(
            self.username, self.password, self.host, self.port,
            self.config.get('db')
        )

    @property
    def fingerprint(self):
        return '[{}://{}:{}]'.format(self.type, self.host, self.port)

    def _check_engine(self):
        if self.engine is None:
            self.logger.error('%s MySQL not connected', self.fingerprint)
            raise MySQLNotConnectedException('MySQL not connected')

    @property
    async def conn(self):
        self._check_engine()

        def coro_finished(f: asyncio.Future):
            if f.cancelled():
                self.logger.warning('acquiring connection is cancelled')
                return
            e = f.exception()
            if e is not None and not isinstance(e,
                                                pymysql.err.OperationalError):
                self.logger.exception(
                    'Exception happened while acquiring connection: %s',
                    str(e), exc_info=e)

        coro = asyncio.ensure_future(self.engine.acquire())
        coro.add_done_callback(coro_finished)
        try:
            res = await asyncio.wait_for(asyncio.shield(coro),
                                         self.request_timeout,
                                         loop=self.loop)
        except asyncio.futures.TimeoutError as e:
            raise MySQLTimeoutError('Timeout error') from e
        return res

    def get_conn(self, *args, **kwargs):
        return self.conn

    def acquire(self):
        return MySQLConnectionCoroContextManager(self, self.conn)

    def release(self, conn):
        self._check_engine()
        return self.engine.release(conn)

    @mysql_connected_check
    def __iter__(self):
        conn = yield from self.conn
        return MySQLConnectionContextManager(self, conn)

    @supply_mysql_conn
    @mysql_connected_check
    async def execute(self, query, *multiparams, **params):
        """
            Используется для неявной передачи conn в одном месте
        """
        conn = params.pop('conn')
        if conn is None:
            raise MySQLNotConnectedException('MySQL not connected')

        def coro_finished(f: asyncio.Future):
            if f.cancelled():
                self.logger.warning(
                    'execute of query is cancelled. q: %s [%s, %s]',
                    query, multiparams, params)
                return
            e = f.exception()
            if e is not None and not isinstance(e,
                                                pymysql.err.OperationalError):
                self.logger.error(
                    'Exception happened while '
                    'executing query \'%s\': %s', query, str(e),
                    exc_info=e)

        coro = asyncio.ensure_future(
            conn.execute(query, *multiparams, **params),
            loop=self.loop
        )
        coro.add_done_callback(coro_finished)
        try:
            res = await coro
            return res
        except asyncio.futures.TimeoutError as e:
            raise MySQLTimeoutError('Timeout error') from e
        except pymysql.InternalError as e:
            err_code, err_msg = e.args
            if err_code == MySQLError.DEADLOCK:
                self.logger.error(
                    'Deadlock happened while executing query %s: %s',
                    query, e.args)
                raise DeadlockError() from e
            raise e

    @mysql_connected_check
    async def begin(self, conn, *args, **kwargs):
        if conn is None:
            raise MySQLNotConnectedException('MySQL not connected')

        def coro_finished(f: asyncio.Future):
            if f.cancelled():
                self.logger.warning('begin of trx is cancelled')
                return
            e = f.exception()
            if e is not None and not isinstance(e,
                                                pymysql.err.OperationalError):
                self.logger.error(
                    'Exception happened while beginning trx: %s',
                    str(e), exc_info=e)

        coro = asyncio.ensure_future(conn.begin(*args, **kwargs),
                                     loop=self.loop)
        coro.add_done_callback(coro_finished)
        try:
            res = await coro
            return res
        except asyncio.futures.TimeoutError as e:
            raise MySQLTimeoutError('Timeout error') from e

    @mysql_connected_check
    async def commit(self, trx, *args, **kwargs):
        if trx is None:
            raise ValueError('trx must be not None')

        def coro_finished(f: asyncio.Future):
            if f.cancelled():
                self.logger.warning('commit of trx is cancelled')
                return
            e = f.exception()
            if e is not None and not isinstance(e,
                                                pymysql.err.OperationalError):
                self.logger.error(
                    'Exception happened while committing trx: %s', str(e),
                    exc_info=e)

        coro = asyncio.ensure_future(trx.commit(*args, **kwargs),
                                     loop=self.loop)
        coro.add_done_callback(coro_finished)
        try:
            res = await coro
        except asyncio.futures.TimeoutError as e:
            raise MySQLTimeoutError('Timeout error') from e

        return res

    @mysql_connected_check
    async def rollback(self, trx, *args, **kwargs):
        if trx is None:
            raise ValueError('trx must be not None')

        def coro_finished(f: asyncio.Future):
            if f.cancelled():
                self.logger.warning('rollback of trx is cancelled')
                return
            e = f.exception()
            if e is not None and not isinstance(e,
                                                pymysql.err.OperationalError):
                self.logger.error(
                    'Exception happened while running rollback trx: %s',
                    str(e), exc_info=e)

        coro = asyncio.ensure_future(trx.rollback(*args, **kwargs),
                                     loop=self.loop)
        coro.add_done_callback(coro_finished)
        try:
            res = await coro
        except asyncio.futures.TimeoutError as e:
            raise MySQLTimeoutError('Timeout error') from e

        return res

    @supply_persist_conn_trx
    @mysql_connected_check
    async def execute_trx(self, query, *multiparams, **params):
        """
            Используется для неявной передачи conn в одном месте
        """
        return await self.execute(query, *multiparams, **params)

    @supply_mysql_conn
    @mysql_connected_check
    async def fetch_all(self, query, *multiparams, **params):
        res = await self.execute(query, *multiparams, **params)
        return await res.fetchall()

    @supply_mysql_conn
    @mysql_connected_check
    async def fetch_by_chunks(self, query, *multiparams, **params):
        chunks_size = params.pop('chunks_size', 50)
        res = await self.execute(query, *multiparams, **params)
        res_list = []
        res_list_part = await res.fetchmany(size=chunks_size)
        while len(res_list_part) > 0:
            res_list += res_list_part
            res_list_part = await res.fetchmany(size=chunks_size)
        return res_list

    @supply_mysql_conn
    @mysql_connected_check
    async def fetch_first(self, query, *multiparams, **params):
        res = await self.execute(query, *multiparams, **params)
        return await res.first()

    @staticmethod
    def dump_sql(func, bind=False):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            out = StringIO()

            def dump(sql, *multiparams, **params):
                out.write('{};\n'.format(
                    str(sql.compile(dialect=dump.dialect)).strip()))

            engine = sa.create_engine('mysql://', strategy='mock',
                                      executor=dump)
            dump.dialect = engine.dialect

            if bind:
                func(*args, bind=engine, **kwargs)
            else:
                func(engine, *args, **kwargs)

            return out.getvalue()

        return func_wrapper



import asyncio
import enum
import functools

import pymysql

from aiokts.store.base_accessors import BaseAccessorException


class MySQLAccessorException(BaseAccessorException):
    pass


class MySQLNotConnectedException(MySQLAccessorException):
    pass


class MySQLTimeoutError(MySQLAccessorException):
    pass


class MySQLError:
    DUPLICATE_KEY = 1062
    DEADLOCK = 1213


class DeadlockError(Exception):
    """
        Это исключение кидается execute'ом в случае возникновения дедлока, чтобы
        вызывающий код мог порестартить транзакцию
    """
    pass


class MySQLConnectionContextManager:
    """Context manager.

    This enables the following idiom for acquiring and releasing a
    connection around a block:

        with (yield from engine) as conn:
            cur = yield from conn.cursor()

    while failing loudly when accidentally using:

        with engine:
            <block>
    """

    __slots__ = ('_connector', '_conn')

    def __init__(self, connector, conn):
        self._connector = connector
        self._conn = conn

    def __enter__(self):
        assert self._conn is not None
        return self._conn

    def __exit__(self, *args):
        try:
            self._connector.release(self._conn)
        finally:
            self._connector = None
            self._conn = None

    # if PY_35:  # pragma: no branch
    #     @asyncio.coroutine
    #     def __aenter__(self):
    #         self._conn = yield from self._coro
    #         return self._conn
    #
    #     @asyncio.coroutine
    #     def __aexit__(self, exc_type, exc, tb):
    #         try:
    #             yield from self._pool.release(self._conn)
    #         finally:
    #             self._pool = None
    #             self._conn = None


class MySQLConnectionCoroContextManager:
    """Context manager.

    This enables the following idiom for acquiring and releasing a
    connection around a block:

        with (yield from engine) as conn:
            cur = yield from conn.cursor()

    while failing loudly when accidentally using:

        with engine:
            <block>
    """

    __slots__ = ('_connector', '_conn', '_coro')

    def __init__(self, connector, coro):
        self._connector = connector
        self._coro = coro
        self._conn = None

    async def __aenter__(self):
        self._conn = await self._coro
        return self._conn

    async def __aexit__(self, exc_type, exc, tb):
        try:
            self._connector.release(self._conn)
        finally:
            self._connector = None
            self._conn = None


def mysql_connected_check(f):
    @functools.wraps(f)
    async def wrap(self, *args, **kwargs):
        try:
            return await f(self, *args, **kwargs)
        except pymysql.err.OperationalError as e:
            self.logger.error('%s Cannot connect to MySQL: %s',
                              self.fingerprint, str(e))
            raise MySQLNotConnectedException(str(e)) from e

    return wrap


def supply_mysql_conn(method):
    @functools.wraps(method)
    async def wrap(self, *args, **kwargs):
        conn = kwargs.get('conn')
        self_conn = conn is None

        if hasattr(self, 'persist'):
            connector = self.persist
        elif hasattr(self, 'get_conn'):
            connector = self
        else:
            raise AttributeError('No connector found')

        try:
            if self_conn:
                conn = await connector.get_conn(mode=Mode.any)
                kwargs['conn'] = conn
            try:
                if asyncio.iscoroutinefunction(method):
                    return await method(self, *args, **kwargs)
                else:
                    return await asyncio.coroutine(method)(
                        self, *args, **kwargs)
            finally:
                if self_conn:
                    connector.release(conn)
        except pymysql.err.OperationalError as e:
            self.logger.error('%s Cannot connect to MySQL: %s',
                              connector.fingerprint, str(e))
            raise MySQLAccessorException(str(e)) from e

    return asyncio.coroutine(wrap)


def supply_persist_conn_trx(method):
    @functools.wraps(method)
    async def wrap(self, *args, **kwargs):
        conn = kwargs.get('conn')
        self_conn = conn is None

        if hasattr(self, 'persist'):
            connector = self.persist
        elif hasattr(self, 'get_conn'):
            connector = self
        else:
            raise AttributeError('No connector found')

        try:
            max_attempts = 3
            for attempt in range(1, max_attempts + 1):
                if attempt > 1:
                    self.logger.warning('Retry #%s transaction for func=%s',
                                        attempt - 1, method)

                trans = None
                if self_conn:
                    conn = await connector.get_conn(mode=Mode.rw)

                    conn.__connector__ = connector
                    trans = await connector.begin(conn=conn)
                    kwargs['conn'] = conn
                try:
                    if asyncio.iscoroutinefunction(method):
                        result = await method(self, *args, **kwargs)
                    else:
                        result = await asyncio.coroutine(method)(
                            self, *args, **kwargs)
                except DeadlockError:
                    if self_conn and conn.in_transaction:
                        await connector.rollback(trans)
                    if not self_conn:
                        raise
                    if attempt >= max_attempts:
                        raise
                except Exception as e:
                    if self_conn and conn.in_transaction:
                        await connector.rollback(trans)
                    raise e
                else:
                    if self_conn and conn.in_transaction:
                        await connector.commit(trans)
                    return result
                finally:
                    if self_conn and not getattr(conn, '__released__', False):
                        connector.release(conn)

        except pymysql.err.OperationalError as e:
            self.logger.error('%s OperationalError: %s',
                              connector.fingerprint, str(e))
            raise MySQLAccessorException(str(e)) from e

    return asyncio.coroutine(wrap)


class Mode(enum.Enum):
    rw = 'rw'
    ro = 'ro'
    any = 'any'


@asyncio.coroutine
def persist_conn_commit(conn):
    if conn is not None and conn.in_transaction:
        if hasattr(conn, '__connector__'):
            connector = conn.__connector__
            if connector is not None:
                trx = conn._transaction
                yield from connector.commit(trx)


@asyncio.coroutine
def persist_conn_rollback(conn):
    if conn is not None and conn.in_transaction:
        if hasattr(conn, '__connector__'):
            connector = conn.__connector__
            if connector is not None:
                trx = conn._transaction
                yield from connector.rollback(trx)


def persist_conn_release(conn):
    if conn is not None:
        connector = conn.__connector__
        if connector is not None:
            connector.release(conn)
            conn.__released__ = True



import asyncpg
import asyncpgsa
from aiokts.store.base_accessors import BaseAccessor


class BasePgAccessor(BaseAccessor):
    DEFAULT_PORT = 5432

    def __init__(self, config, type, store, loop=None):
        super().__init__(config, type, store, loop=loop)
        self._pool = None

    @property
    def db_name(self):
        return self.config['db']

    async def acquire(self):
        return await self._pool.acquire()

    async def release(self, conn):
        return await self._pool.release(conn)

    async def _connect(self):
        pool_min_size = self.config.get('pool_min', 1)
        pool_max_size = self.config.get('pool_max', 1)
        self._pool = await asyncpg.create_pool(host=self.host,
                                               port=self.port,
                                               user=self.username,
                                               password=self.password,
                                               database=self.db_name,
                                               loop=self.loop,
                                               min_size=pool_min_size,
                                               max_size=pool_max_size)

    async def _disconnect(self):
        return await self._pool.close()

    def compile_q(self, q):
        return asyncpgsa.compile_query(q)

    async def _execute_operation(self, operation, query, conn=None,
                                 *args, **kwargs):
        q, q_args = self.compile_q(query)
        if conn:
            return await getattr(conn, operation)(str(q), *q_args)
        else:
            async with self._pool.acquire() as conn:
                return await getattr(conn, operation)(str(q), *q_args)

    async def execute(self, q, conn=None, *args, **kwargs):
        return await self._execute_operation("execute", q, conn,
                                             *args, **kwargs)

    async def fetchrow(self, q, conn=None, *args, **kwargs):
        return await self._execute_operation("fetchrow", q, conn,
                                             *args, **kwargs)

    async def exists(self, q, conn=None, *args, **kwargs):
        return bool(await self.fetch(q, conn, *args, **kwargs))

    async def fetch(self, q, conn=None, *args, **kwargs):
        return await self._execute_operation("fetch", q, conn,
                                             *args, **kwargs)

    @staticmethod
    def transaction(conn, **kwargs):
        return conn.transaction(**kwargs)



import argparse
import inspect
import logging
import logging.config
import os
import pkgutil
import sys

from aiokts.managecommands import Command
from aiokts.store import Store

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(CURRENT_DIR)


class BaseManage(object):
    commands_package_path = None
    store_cls = Store
    _modules = {}
    _commands = None

    def __init__(self):
        self._config_dir = None
        self._config_path = None
        self._config = None
        self._logger = None

        assert self.commands_package_path is not None, \
            'Must specify path to where commands are'
        self.commands_package_path = os.path.abspath(
            os.path.join(
                os.path.dirname(inspect.getfile(self.__class__)),
                self.commands_package_path))
        self.logger.debug('Commands path: %s', self.commands_package_path)

    @property
    def commands(self):
        if self._commands is None:
            self._commands = ['help']
            for loader, name, ispkg in \
                    pkgutil.iter_modules([self.commands_package_path]):
                if not ispkg:
                    self._commands.append(name)
                    self._modules[name] = loader.find_module(name)
        return self._commands

    @property
    def config(self):
        return {}

    def help(self):
        print('Available commands:\n - %s' % ('\n - '.join(self.commands)))

    def run(self):
        args = self._parse_manage_arguments()

        command = None
        try:
            command = args.command
            if command not in self.commands:
                logging.error('Command %s not found' % command)
                self.help()
                return 1

            if command == 'help':
                self.help()
                return 0

            self._run_command(command, *args.opts)
        except Exception:
            self.logger.exception('Exception while running command %s',
                                  command)
            return 2
        except BaseException:
            self.logger.exception('Exception while running command %s',
                                  command)
            return 3

    def _run_command(self, command, *args):
        module = self._modules[command].load_module(command)
        if hasattr(module, 'main'):
            module.main(*args)

        cmd_cls = None
        for name, cls in module.__dict__.items():
            if isinstance(cls, type) and issubclass(cls, Command)\
                    and cls.__module__ == module.__name__:
                cmd_cls = cls
                break

        assert cmd_cls is not None, \
            "Couldn't find Command in command {}".format(command)
        cmd = cmd_cls(self)
        cmd.run(*args)

    def _parse_manage_arguments(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('command', help='command to execute')
        parser.add_argument('opts', nargs=argparse.REMAINDER, default=None)
        args = parser.parse_args()

        return args

    @property
    def logger(self):
        if self._logger is None:
            self._logger = logging.getLogger('Manager')
        return self._logger


def main(manager_cls):
    manage = manager_cls()
    exit(manage.run())


if __name__ == '__main__':
    main(Manage)



import argparse
import collections
import copy
import os

import sys

import logging
import logging.config
from logging.handlers import RotatingFileHandler

import yaml

__all__ = (
    'Settings',
)


def update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


class Settings:
    def __init__(self, default_config_dir=None, local_config_dir=None, **kwargs):
        self.default_config_dir = default_config_dir
        self.local_config_dir = local_config_dir

        self.config_dir = kwargs.get('config_dir')

        self.config_path = kwargs.get('config_path')
        self.local_config_path = kwargs.get('local_config_path')
        self.logging_path = kwargs.get('logging_path')
        self.log_path = kwargs.get('log_path')

        self.root_config = kwargs.get('root_config')
        self.local_config = kwargs.get('local_config')
        self.config = None

    def parse(self):
        args = self.parse_arguments()

        is_config_by_args = args.config_dir is not None
        self.config_dir = args.config_dir if is_config_by_args else self.default_config_dir

        if self.config_path is None or is_config_by_args:
            self.config_path = os.path.join(self.config_dir, 'main', 'main.yaml')

        self.logging_path = args.logging_config if args.logging_config else self.logging_path
        if self.logging_path is None and self.config_dir:
            self.logging_path = os.path.join(self.config_dir, 'api_hour', 'logging.ini')

        self.root_config = self.parse_config(self.config_path)

        if not self.local_config_path and self.local_config_dir:
            self.local_config_path = os.path.join(self.local_config_dir, 'main', 'main.yaml')
        if self.local_config_path and os.path.exists(self.local_config_path):
            self.local_config = self.parse_config(self.local_config_path)

        self.config = copy.deepcopy(self.root_config)
        if self.local_config is not None:
            self.config = update(self.config, self.local_config)

        self.log_path = args.log

    def parse_arguments(self):
        parser = argparse.ArgumentParser()

        parser.add_argument(
            '--config_dir',
            type=str,
            help='config dir path',
            default=None
        )
        parser.add_argument(
            '--logging_config',
            type=str,
            help='logging config path',
            default=None
        )
        parser.add_argument(
            '--log',
            type=str,
            help='log file path',
            default=None
        )

        args, left = parser.parse_known_args()
        sys.argv = sys.argv[:1] + left

        if args.config_dir is not None:
            assert os.path.isdir(args.config_dir), \
                'config_dir `{}` does not exist'.format(args.config_dir)

        if args.logging_config is not None:
            assert os.path.exists(args.logging_config), \
                'logging_config `{}` does not exist'.format(
                    args.logging_config)

        return args

    def parse_config(self, path=None):
        if path is None:
            path = self.config_path
        with open(path, 'r') as f:
            return yaml.load(f)

    def setup_logging(self):
        if self.logging_path:
            logging.config.fileConfig(
                self.logging_path,
                disable_existing_loggers=False
            )

        if self.log_path is not None:
            old_file_handler = None
            for handler in logging.root.handlers[:]:
                if isinstance(handler, logging.FileHandler):
                    old_file_handler = handler
                    logging.root.removeHandler(handler)
            if old_file_handler is not None:
                file_handler = RotatingFileHandler(self.log_path, 'w+', 104857600, 100)
                file_handler.setFormatter(old_file_handler.formatter)
                logging.root.addHandler(file_handler)

    def __getitem__(self, name):
        return self.config[name]

    def get(self, *args, **kwargs):
        try:
            result = self
            for key in args:
                result = result[key]
        except:
            for key in ['default', 'fallback']:
                if key in kwargs:
                    return kwargs[key]
            return None

        return result

    def getint(self, *args, **kwargs):
        return self.get(*args, **kwargs)

    def getfloat(self, *args, **kwargs):
        return self.get(*args, **kwargs)

    def getboolean(self, *args, **kwargs):
        return self.get(*args, **kwargs)




import asyncio

import logging


class StoreException(Exception):
    pass


class Store(object):
    def __init__(self, config, need=None, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        self.logger = logging.getLogger('store')
        self.config = self.check_config(config)
        self._need = None
        self.need(need)

        self._accessors = None
        self.init_store()

        self.__connect_coros = []

    def need(self, need=None):
        if need is not None:
            self._need = set(need) if need is not None else set()
        return self._need

    def get_extra_location(self):
        return []

    def init_store(self):
        from aiokts.store import base_accessors

        self._accessors = {}

        # Importing BaseAccessor class
        base_accessor_class = vars(base_accessors).get('BaseAccessor')

        # Importing all modules required in self._need and collecting map
        # (module_name -> accessor_type)
        if self._need is None:
            return

        for accessor_type in self._need:
            accessor_type_found = None
            locations = [base_accessors] + self.get_extra_location()
            for location in locations:
                if accessor_type_found is True:
                    break
                module_name = '{}.{}'.format(location.__name__, accessor_type)
                try:
                    module = __import__(module_name, fromlist='dummy')
                    accessor_type_found = True
                except ImportError as e:
                    if module_name in str(e):
                        accessor_type_found = False
                        continue
                    else:
                        raise

                registered_accessors = set()
                for accessor_class in \
                        self._classes_in_module(module, base_accessor_class):
                    if accessor_type not in self._accessors \
                            and accessor_class not in registered_accessors:
                        try:
                            conf = self.config.get(accessor_type)
                            self._accessors[accessor_type] = \
                                accessor_class(config=conf,
                                               type=accessor_type,
                                               store=self,
                                               loop=self.loop)
                            registered_accessors.add(accessor_class)
                        except Exception as e:
                            self.logger.error('Error with %s accessor: %s',
                                              accessor_type, str(e))
                            raise e

            if not accessor_type_found:
                raise StoreException(
                    'Accessor with type {} not found'.format(accessor_type))

    def _classes_in_module(self, module, base):
        md = module.__dict__
        classes = []
        for _, c in module.__dict__.items():
            if isinstance(c, type) \
                    and c.__module__ == module.__name__ \
                    and issubclass(c, base) \
                    and c is not base:
                classes.append(c)
        return classes

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if name in self._accessors:
                return self._accessors[name]
            else:
                raise StoreException(
                    'Accessor \'{}\' not found in Store. '
                    'Probably not specified in need'.format(name))

    async def wait_connected(self):
        coros = []
        for accessor_type, c in self._accessors.items():
            coro = asyncio.ensure_future(c.wait_connected(), loop=self.loop)
            coros.append(coro)

        if len(coros) > 0:
            await asyncio.wait(coros, loop=self.loop)

    def check_config(self, config):
        return config

    async def connect(self):
        try:
            all_success = True

            def on_all_connected(f):
                nonlocal all_success

                if all_success:
                    self.logger.info('Connected to all')
                else:
                    self.logger.info('Connected to all with errors')

                self.__connect_coros = None

            def on_one_connected(f):
                nonlocal all_success

                cancelled = f.cancelled()
                exc = f.exception()
                if not cancelled and not exc:
                    return

                f_accessor_type = getattr(f, '_accessor_type', 'UNK_TYPE')

                if cancelled:
                    self.logger.error(
                        'Connection to [%s] is cancelled', f_accessor_type)
                    all_success = False
                elif exc:
                    self.logger.error(
                        'Exception while connecting to [%s] : %s',
                        f_accessor_type, str(exc), exc_info=exc)
                    all_success = False

            coros = []

            if self._accessors:
                for accessor_type, c in self._accessors.items():
                    coro = asyncio.ensure_future(c.connect(), loop=self.loop)
                    coro.add_done_callback(on_one_connected)
                    coro._accessor_type = accessor_type

                    coros.append(coro)

            if not coros:
                self.logger.info('No connectors to connect to')
                return

            self.__connect_coros = coros

            wait_coro = asyncio.ensure_future(
                asyncio.wait(self.__connect_coros,
                             loop=self.loop,
                             return_when=asyncio.ALL_COMPLETED),
                loop=self.loop
            )
            wait_coro.add_done_callback(on_all_connected)
            await wait_coro
        except asyncio.CancelledError:
            pass
        except Exception as e:
            self.logger.exception('Error while connecting')

    async def disconnect(self):
        try:
            # all_success = True
            #
            # def on_all_disconnected(f):
            #     nonlocal all_success
            #
            #     if all_success:
            #         self.logger.info('Disconnected from all')
            #     else:
            #         self.logger.info('Disconnected from all with errors')
            #
            # def on_one_disconnected(f):
            #     cancelled = f.cancelled()
            #     exc = f.exception()
            #     if not cancelled and not exc:
            #         return
            #
            #     f_accessor_type = getattr(f, '_accessor_type', 'UNK_TYPE')
            #
            #     if cancelled:
            #         self.logger.error('Disconnection from [%s] is cancelled',
            #                           f_accessor_type)
            #     elif exc:
            #         self.logger.error(
            #             'Exception while disconnecting from [%s]: %s',
            #             f_accessor_type, str(exc), exc_info=exc
            #         )

            # For some funky reason, asyncio.wait crashes Daemon on exit
            # coros = set()
            if self._accessors:
                for accessor_type, c in self._accessors.items():
                    await c.disconnect()
                    # coro = c.disconnect()
                    # coro.add_done_callback(on_one_disconnected)
                    # coro._accessor_type = accessor_type
                    # coros.add(coro)

            # if not coros:
            #     self.logger.info('No connectors to disconnect from')
            #     return

            # wait_coro = asyncio.ensure_future(
            #     asyncio.wait(coros,
            #                  loop=self.loop,
            #                  return_when=asyncio.ALL_COMPLETED),
            #     loop=self.loop
            # )
            # wait_coro.add_done_callback(on_all_disconnected)
        except Exception as e:
            self.logger.error(e)



import asyncio

import motor.motor_asyncio
from pymongo.errors import AutoReconnect, ConnectionFailure

from aiokts.store.base_accessors import BaseAccessor, ConfigurationError


class BaseMongoDbAccessor(BaseAccessor):
    DEFAULT_PORT = 27017

    def __init__(self, config, type, store, loop=None):
        super().__init__(config, type, store, loop=loop)
        self._conn = None

    async def _connect(self):
        self._conn = motor.motor_asyncio.AsyncIOMotorClient(
            self._build_connection_string(),
            io_loop=self.loop
        )
        try:
            await self.wait_db()
        except AutoReconnect as e:
            self.logger.error("Couldn't connect to db %s", self.fingerprint)
            await self.wait_db()

    async def _disconnect(self):
        if self._conn is not None:
            self._conn = None

    @property
    def db_name(self):
        return self.config.get('db')

    @property
    def db(self):
        return self._conn[self.db_name] if self._conn is not None else None

    async def ping(self):
        try:
            await self._conn.admin.command({'ping': 1})
            return True
        except ConnectionFailure:
            return False

    async def wait_db(self):
        pong = False
        while not pong:
            pong = await self.ping()
            if not pong:
                self.logger.warning('%s is unavailable. Waiting.',
                                    self.fingerprint)
                await asyncio.sleep(1.0, loop=self.loop)
        return self.db

    def check_config(self):
        super().check_config()
        if self.db_name is None:
            raise ConfigurationError('db is required for MongoDbConnector')

    def _build_connection_string(self):
        if self.has_credentials:
            s = 'mongodb://%s:%s@%s:%d/%s'
            args = (self.username, self.password, self.host, self.port,
                    self.db_name)
        else:
            s = 'mongodb://%s:%d/%s'
            args = (self.host, self.port, self.db_name)
        s %= args
        return s


async def mongo_wait_connected_on_coro(coro, *args, **kwargs):
    retry = 0
    while True:
        retry += 1
        try:
            res = await coro(*args, **kwargs)
            if retry > 1:
                print('Restored mongo connection in {}'.format(coro))
            return res
        except AutoReconnect:
            print('Waiting for mongo connection in {}'.format(coro))
            await asyncio.sleep(1.0)
